from typing import Union

from confluence_poster.main_helpers import StateConfig
from confluence_poster.convert_utils import get_representation_for_format
from confluence_poster.page_location_helpers import determine_location
from confluence_poster.poster_config import Page


class CreationResult:
    def __init__(
        self,
        page_created: bool,
        page_id: Union[int, None] = None,
        comment: Union[str, None] = None,
    ):
        self.page_created = page_created
        self.page_id = page_id
        self.comment = comment

    def __bool__(self):
        return self.page_created


def create_page(page: Page, state: StateConfig, create_in_root: bool) -> CreationResult:
    """Handles user input for page creation.

    :return CreationResult that contains info on whether the page was created and its ID
    """
    echo = state.print_function
    confirm = state.confirm_function

    if state.force_create or confirm("Should the page be created?", default=True):
        if location := determine_location(
            page=page, create_in_root=create_in_root, state=state
        ):
            echo("Creating page...")
            page_id = state.confluence_instance.create_page(
                space=page.page_space,
                title=page.page_title,
                body=page.page_text,
                parent_id=location.parent_page_id,
                representation=get_representation_for_format(
                    page.page_file_format
                ).value,
            )["id"]
            if location.parent_page_id is None:
                page_location_msg = f"in root of the space '{page.page_space}'"
            else:
                page_location_msg = f"under page #{location.parent_page_id}, '{location.parent_page_name}'"

            echo(
                f"Created page #{page_id} {page_location_msg} called '{page.page_title}'."
            )
            return CreationResult(True, page_id)
        else:
            return CreationResult(
                False, comment="Could not determine location for the page."
            )
    else:
        return CreationResult(False, comment="User cancelled creation when prompted.")



import typer
from click import Choice
from pathlib import Path
from functools import reduce
from typing import Union, Any, Tuple, FrozenSet, Iterable, Callable
from tomlkit import document, parse, table, dumps
from tomlkit.parser import TOMLDocument
from tomlkit.items import Table
from dataclasses import dataclass
from copy import deepcopy

from confluence_poster.poster_config import AllowedFileFormat


@dataclass
class DialogParameter:
    """This class serves as a wrapper around title(str) with extra parameters to be consumed by the dialog prompt"""

    title: str
    comment: Union[str, None] = None  # for passing a comment
    type: Union[type, None, Choice] = None  # for passing type to input
    required: bool = True
    hide_input: bool = False

    def __eq__(self, other):
        """Method to help looking up members of this class when the dialog runs"""
        if isinstance(other, DialogParameter):
            return self.title == other.title
        elif isinstance(other, str):
            return self.title == other
        else:
            raise ValueError

    def __hash__(self):
        """To be used very carefully"""
        return hash(self.title)

    def __getattr__(self, item):
        """To proxy """

        def _missing(*args, **kwargs):
            method = getattr(self.title, item)
            return method(*args, **kwargs)

        return _missing

    def __str__(self):
        return self.title


def _dialog_prompt(parameter: Union[DialogParameter, str], default_value=None) -> Any:
    """Provides a single prompt of the dialog.

    Special case: if None is returned - then the user skipped this parameter"""
    _default_value = default_value
    if type(parameter) is not DialogParameter:
        # For handling default case without special information
        parameter = DialogParameter(title=parameter)

    message = [f"Please provide a value for {parameter.title}."]
    if parameter.comment is not None:
        message += [f"Comment: {parameter.comment}"]
    if not parameter.required:
        message += ["This parameter is optional. Press [Enter] to skip it."]
        if default_value is None:
            _default_value = ""

    if parameter.hide_input:
        if default_value is not None:
            message += [
                f"Current value is set, but input is hidden, indicating a sensitive field.",
                "Press [Enter] to reuse the current value.",
            ]
        else:
            message += ["This parameter is marked as sensitive, input is hidden"]
    elif default_value is not None:
        message += [f"Current value is {default_value}. Press [Enter] to use it."]

    new_value = typer.prompt(
        text="\n".join(message),
        default=_default_value,
        type=parameter.type,
        hide_input=parameter.hide_input,
        show_default=not parameter.hide_input,
    )
    if not parameter.required and default_value is None and new_value == "":
        new_value = None

    return new_value


def _get_attribute_by_path(attribute_path: str, config: TOMLDocument) -> Any:
    """Given attribute path like path1.path2.attribute return value of attribute, None if attribute is empty
    or path does not exist"""
    attribute_path = attribute_path.split(".")
    try:
        return reduce(lambda caret, key: caret.get(key, None), attribute_path, config)
    except AttributeError:  # for cases when we try to call get() on None
        return None


def _get_filled_attributes(config: TOMLDocument) -> Tuple[str]:
    """For a config like this:

    foo = 'bar'
    [table]
    foo = 'bar'
    [table.child]
    foo = 'bar'

    produces (foo, table.foo, table.child.foo)
    """

    def key_walk(d: dict, base_path: str = "") -> str:
        """A generator that traverses into a TOMLDocument, keeping track through base_path"""
        for key in d:
            path = base_path
            if type(d[key]) is Table:
                path += key + "."
                yield from key_walk(d[key], base_path=path)
            else:
                yield path + key

    result = tuple(key_walk(config))
    return result


def _create_or_update_attribute(
    attribute: str, config: TOMLDocument, value: str
) -> TOMLDocument:
    """Given attribute path path1.path2.attribute, do the following:

    1. Traverse the document, creating tables for path1 and path2 if they do not exist. 'path2' would be nested in path1
    2. Set the attribute attribute to the value

    Returns a copy of config with updated value
    """
    *attribute_path, attribute_name = attribute.split(".")
    _config = deepcopy(config)
    caret = _config
    for path_node in attribute_path:  # more clear than a reduce() call
        next_node = caret.get(path_node)
        if next_node is None:
            caret[path_node] = table()
        caret = caret.get(path_node)

    caret[attribute_name] = value
    return parse(dumps(_config))


def print_config_with_hidden_attrs(
    config: TOMLDocument, hidden_attributes: Iterable[Union[str, DialogParameter]]
) -> str:
    """Given a path on the filesystem and a list of hidden attributes, returns content of that file,
    redacting the sensitive fields"""
    _config = deepcopy(config)
    for attribute in hidden_attributes:
        if _get_attribute_by_path(str(attribute), _config) is not None:
            _config = _create_or_update_attribute(
                str(attribute), _config, value="[REDACTED]"
            )

    return dumps(_config)


def config_dialog(
    filename: Union[Path, str],
    attributes: Iterable[Union[str, DialogParameter]],
    config_print_function: Callable = lambda _: print(_),
    incremental: bool = False,
) -> Union[None, bool]:
    """Checks if filename exists and goes through the list of attributes asking the user for the values

    :param filename: filename (path or string) containing the config to be output
    :param attributes: list of parameter paths or DialogParameters to be displayed
    :param config_print_function: function that prints the config file.
    Can be overridden using print_config_file function to preserve list of redacted attributes.
    :param incremental: if set to True - suppresses the prompt to overwrite the file
    """
    if type(filename) is str:
        filename = Path(filename)
    new_config = document()
    if filename.exists():
        typer.echo(f"File {filename} already exists.")
        new_config = parse(filename.read_text())
        if not incremental:
            typer.echo("Current content:")
            typer.echo(config_print_function(new_config))
            if not typer.confirm(f"File {filename} exists. Overwrite?", default=False):
                return  # do not save this config file

    # Process attributes list
    for attr in attributes:
        current_value = _get_attribute_by_path(attr, new_config)
        if current_value is not None:
            if incremental:
                raise Exception(
                    f"Incremental is set, but there is already a value for {attr}. This is probably a bug."
                )
            if not typer.confirm(
                f"Would you like to overwrite current value of {attr}: {current_value}?",
                default=True,
            ):
                continue  # next attribute
        new_value = _dialog_prompt(parameter=attr, default_value=current_value)

        if new_value is not None:
            new_config = _create_or_update_attribute(
                attribute=attr, config=new_config, value=new_value
            )

    typer.echo(f"Config to be saved:")
    typer.echo(config_print_function(new_config))

    save = typer.confirm(
        f"Would you like to save it as {filename}? "
        "The wizard will create all missing parent directories",
        default=True,
    )
    if save:
        typer.echo(f"Saving config as {filename}")
        filename.parent.mkdir(parents=True, exist_ok=True)
        filename.write_text(dumps(new_config))
        if any([_.hide_input for _ in attributes if isinstance(_, DialogParameter)]):
            typer.echo(
                "Since a sensitive parameter was passed - saving the config file with 600 permissions."
            )
            filename.chmod(0o600)
        return True
    else:
        return False


def get_filled_attributes_from_file(filename: Union[Path, str]) -> FrozenSet[str]:
    if type(filename) is str:
        filename = Path(filename)
    if not filename.exists():
        return frozenset()
    return frozenset(_get_filled_attributes(parse(filename.read_text())))


def _generate_next_page(filename: Union[Path, str]) -> int:
    existing_pages = set(
        map(
            lambda a: a.split(".")[1],
            filter(
                lambda _: _.startswith("pages")
                and not _.startswith("pages.default")
                and not _ == "pages",
                get_filled_attributes_from_file(filename),
            ),
        )
    )
    page_number = 1
    while True:
        if f"page{page_number}" not in existing_pages:
            return page_number
        else:
            page_number = page_number + 1


def generate_page_dialog_params(
    page_no: int,
) -> Tuple[DialogParameter, DialogParameter, DialogParameter, DialogParameter]:
    # noinspection PyUnresolvedReferences
    return (
        DialogParameter(
            title=f"pages.page{page_no}.page_title", comment="The title of the page"
        ),
        DialogParameter(
            title=f"pages.page{page_no}.page_file", comment="File containing page text"
        ),
        DialogParameter(
            title=f"pages.page{page_no}.page_file_format",
            comment="Text format of the page file. None or default - the script will try to guess it at runtime.",
            type=Choice(
                [_[0] for _ in AllowedFileFormat.__members__.items()]
            ),  # making the linter happy, see JetBrains PY-36205
            required=False,
        ),
        DialogParameter(
            title=f"pages.page{page_no}.page_space",
            comment="Key of the space with the page",
            required=False,
        ),
    )


def page_add_dialog(
    filename: Union[Path, str], config_print_function=lambda _: print(_)
) -> bool:
    """Wrapper around config_dialog that generates a new page section"""
    # processes list of pages.page1, pages.page2 to "page1", "page2"
    page_number = _generate_next_page(filename)
    return config_dialog(
        filename,
        attributes=generate_page_dialog_params(page_number),
        config_print_function=config_print_function,
        incremental=True,
    )



from typing import Union

from confluence_poster.main_helpers import StateConfig, get_page_url
from confluence_poster.poster_config import Page


def _find_parent(parent_name: str, space: str, state: StateConfig) -> Union[int, None]:
    """Helper function to locate the parent page.

    :return page id if parent is found, None otherwise
    """
    state.print_function(f"Looking for the parent page with title '{parent_name}'")
    if parent_page := state.confluence_instance.get_page_by_title(
        space=space, title=parent_name, expand=""
    ):
        # according to Atlassian REST API reference, '_links' is a legitimate way to access links
        parent_link = get_page_url(parent_name, space, state.confluence_instance)
        _parent_id = parent_page["id"]
        state.print_function(
            f"Found page #{_parent_id}, called '{parent_name}'. URL is: {parent_link}"
        )
        return _parent_id
    else:
        state.print_function(f"Parent page '{parent_name}' not found")
        return None


def _prompt_for_parent(state: StateConfig) -> str:
    """Function that handles user input """
    prompt = state.prompt_function
    return prompt("Which page should the script look for?")


class LocationResult:
    def __init__(
        self,
        create_page: bool,
        parent_page_id: Union[int, None] = None,
        parent_page_name: Union[str, None] = None,
    ):
        self.create_page = create_page
        self.parent_page_id = parent_page_id
        self.parent_page_name = parent_page_name

    def __bool__(self):
        return self.create_page


def determine_location(
    page: Page,
    create_in_root: bool,
    state: StateConfig,
) -> LocationResult:
    """Handles user input when creating the page

    :return bool - whether page should be created, page_id of parent page if it exists, None for creating in root
    """
    echo = state.print_function
    confirm = state.confirm_function
    always_echo = state.always_print_function

    if create_in_root:
        echo(f"Will create the page in root of space {page.page_space}")
        return LocationResult(True, None)

    if page.parent_page_title:
        echo(
            f"Will create the page under the specified parent page '{page.parent_page_title}'"
        )
        parent_page_id = _find_parent(
            parent_name=page.parent_page_title, space=page.page_space, state=state
        )
        if parent_page_id is not None:
            return LocationResult(True, parent_page_id, page.parent_page_title)
        else:
            always_echo(
                f"Provided page '{page.parent_page_title}' not found in space '{page.page_space}'.\n"
                "Skipping page."
            )
            return LocationResult(False)

    else:
        while confirm(
            f"Should the script look for a parent in space {page.page_space}?"
            f" (N to be prompted to create the page in the space root)\n"
            f"Hint: you can pass --create-in-space-root or --parent-page-title to skip this prompt."
        ):
            parent_title = _prompt_for_parent(state)
            if parent_id := _find_parent(
                parent_name=parent_title, space=page.page_space, state=state
            ):
                if confirm(
                    f"Proceed to create the page '{page.page_title}' under page '{parent_title}'?"
                ):
                    return LocationResult(True, parent_id, parent_title)
                else:
                    return LocationResult(False)
        else:
            if confirm(
                f"Create the page in the root of space '{page.page_space}'? (N will skip the page)"
            ):
                return LocationResult(True, None)
            else:
                return LocationResult(False)



from xdg import xdg_config_dirs, xdg_config_home
from confluence_poster.poster_config import Config, PartialConfig
from collections import UserDict
from collections.abc import Mapping
from pathlib import Path


def merge_configs(first_config: Mapping, other_config: Mapping):
    """Merges two configs together, like so:
    {'auth': {'user': 'a'}} + {'auth': {'password': 'b'}} = {'auth': {'user': 'a', 'password': 'b'}}"""
    for key in set(first_config).union(other_config):
        if key in first_config and key in other_config:
            check_list = [
                isinstance(_, dict) for _ in [first_config[key], other_config[key]]
            ]
            if all(check_list):
                yield key, dict(merge_configs(first_config[key], other_config[key]))
            elif isinstance(first_config[key], type(other_config[key])):
                yield key, other_config[key]
            elif check_list.count(True) == 1:  # one and only one is a dict
                raise ValueError(
                    f"{key} is a section in one of the configs, not in another"
                )
            else:
                raise ValueError(f"{key} key is malformed in one of the configs")

        elif key in first_config:
            yield key, first_config[key]
        else:
            yield key, other_config[key]


def load_config(local_config: Path) -> Config:
    """Function that goes through the config directories trying to load the config.
    Reads configs from XDG_CONFIG_DIRS, then from XDG_CONFIG_HOME, then the local one - either the default one, or
    supplied through command line."""
    final_config = UserDict()
    for path in xdg_config_dirs()[:-1] + [xdg_config_home()]:
        config_path = path / "confluence_poster/config.toml"
        if config_path.exists():
            final_config = dict(
                merge_configs(final_config, PartialConfig(file=config_path))
            )

    final_config = dict(merge_configs(final_config, PartialConfig(file=local_config)))

    return Config(data=final_config)



from typing import Iterable
from pathlib import Path

from confluence_poster.main_helpers import StateConfig, PostedPage


def attach_files_to_page(
    page: PostedPage, files: Iterable[Path], state: StateConfig
) -> None:
    echo = state.print_function
    always_echo = state.always_print_function

    always_echo("Uploading the files")
    for path in files:
        if path.is_file():
            echo(f"\tUploading file {path.name}...")
            state.confluence_instance.attach_file(
                str(path), name=path.name, page_id=str(page.page_id)
            )
            echo(f"\tUploaded file {path.name}.")
    always_echo("Done uploading files")



from enum import Enum
from pathlib import Path
from atlassian import Confluence
from requests import Response
from markdown import markdown

from confluence_poster.poster_config import AllowedFileFormat


def post_to_convert_api(confluence: Confluence, text: str) -> str:
    url = "rest/tinymce/1/markdownxhtmlconverter"
    # the endpoint returns plain text, need to redefine the default header
    headers = {"Content-Type": "application/json"}
    # Keep until https://github.com/atlassian-api/atlassian-python-api/pull/684 is released
    original_advanced_mode = confluence.advanced_mode
    if confluence.advanced_mode is False or confluence.advanced_mode is None:
        confluence.advanced_mode = True

    response: Response = confluence.post(url, data={"wiki": text}, headers=headers)
    # No way to trigger failure for this during tests
    response.raise_for_status()  # pragma: no cover

    confluence.advanced_mode = original_advanced_mode

    return response.text


def convert_using_markdown_lib(text: str) -> str:
    return markdown(text, extensions=("tables", "fenced_code"))


def guess_file_format(page_file: str) -> AllowedFileFormat:
    """Attempts to guess the file format from the page file by the file extension.
    If the extension is unknown raises an error"""
    md_file_formats = {
        ".markdown",
        ".mdown",
        ".mkdn",
        ".md",
        ".mkd",
        ".mdwn",
        ".mdtxt",
        ".mdtext",
        ".text",
        ".Rmd",
    }
    cw_file_formats = {".confluencewiki", ".wiki"}
    html_file_formats = {".html"}
    if (suffix := Path(page_file).suffix) in md_file_formats:
        return AllowedFileFormat.markdown
    elif suffix in cw_file_formats:
        return AllowedFileFormat.confluencewiki
    elif suffix in html_file_formats:
        return AllowedFileFormat.html
    else:
        raise ValueError(f"File format of file {page_file} could not be guessed.")


class Representation(Enum):
    wiki = "wiki"
    editor = "editor"


def get_representation_for_format(file_format: AllowedFileFormat) -> Representation:
    if file_format == AllowedFileFormat.markdown:
        return Representation.editor
    elif file_format == AllowedFileFormat.html:
        return Representation.editor
    elif file_format == AllowedFileFormat.confluencewiki:
        return Representation.wiki
    else:
        raise ValueError(
            f"Could not determine representation value for {file_format}. This is probably a bug."
        )



import toml
from pathlib import Path
from dataclasses import dataclass
from typing import Union
from operator import attrgetter
from itertools import groupby
from collections import UserDict
from marshmallow import Schema, fields, ValidationError
from enum import Enum


class AllowedFileFormat(str, Enum):
    confluencewiki = "confluencewiki"
    markdown = "markdown"
    html = "html"
    none = "None"


@dataclass
class Page:
    page_title: str
    page_file: str
    page_space: Union[str, None]
    parent_page_title: Union[str, None] = None
    _page_text: str = ""

    def __eq__(self, other) -> bool:
        if not isinstance(other, Page):
            raise ValueError
        return (self.page_title == other.page_title) and (
            self.page_space == other.page_space
        )

    @property
    def page_text(self) -> str:
        if self._page_text == "":
            if (_file := Path(self.page_file)).exists():
                self._page_text = Path(self.page_file).read_text()
        return self._page_text

    @page_text.setter
    def page_text(self, value: str):
        self._page_text = value

    page_file_format: AllowedFileFormat = AllowedFileFormat.none
    force_overwrite: Union[bool, None] = False


class AllowedFileFormatField(fields.Field):
    def _deserialize(self, value, attr, data, **kwargs):
        if not isinstance(value, AllowedFileFormat):
            raise ValidationError(f"Invalid value for field: {value}")
        return value.value


class PageSchema(Schema):
    page_title = fields.Str()
    page_file = fields.Str()
    page_space = fields.Str()
    parent_page_title = fields.Str(missing=None)
    _page_text = fields.Str()
    page_file_format = AllowedFileFormatField(
        default=AllowedFileFormat.none, missing=AllowedFileFormat.none
    )
    force_overwrite = fields.Boolean(default=False)


@dataclass
class Auth:
    url: str
    username: str
    password: Union[str, None]
    is_cloud: bool = False


class PartialConfig(UserDict):
    """A class that allows reading file contents or data from a dictionary"""

    def __init__(
        self, file: Union[str, Path, None] = None, data: Union[dict, None] = None
    ):
        if file is None and data is None:
            raise ValueError("No data provided")

        if file is not None and data is not None:
            raise ValueError(
                "Both file and data were provided, cannot determine preference"
            )

        if data is not None:
            _data = data
        else:
            _data = toml.load(file)

        super(PartialConfig, self).__init__(_data)


class Config(PartialConfig):
    """Class that loads the config file and provides interface to its values"""

    def __init__(
        self, file: Union[str, Path, None] = None, data: Union[dict, None] = None
    ):
        """File: file to read config from
        is_global: allows suppressing checks to implement config inheritance"""
        super(Config, self).__init__(file, data)

        _ = self.data
        self.pages = _["pages"]
        self.auth = _["auth"]
        self.author = _.get("author", None)

    @property
    def pages(self):
        return self.__pages

    @pages.setter
    def pages(self, pages: dict):
        self.__pages = list()
        default_space = None
        for item in pages:
            item_content = pages[item]
            if isinstance(item_content, dict):
                if item == "default":
                    default_space = item_content.get(
                        "page_space", None
                    )  # None is OK here, checked later
                    if not isinstance(default_space, str) or default_space is None:
                        raise ValueError("default.page_space should be a string")
                else:  # this is a page definition
                    for prop in item_content:  # TODO: better validation
                        if not isinstance(item_content[prop], str):
                            if prop != "force_overwrite":
                                raise ValueError(
                                    f"{prop} property of a page is not a string"
                                )

                    page = Page(
                        page_title=item_content.get("page_title", None),
                        page_file=item_content["page_file"],
                        page_file_format=AllowedFileFormat(
                            item_content.get("page_file_format", "None")
                        ),
                        page_space=item_content.get("page_space", None),
                        parent_page_title=item_content.get("page_parent_title", None),
                        force_overwrite=item_content.get("force_overwrite", False),
                    )
                    self.__pages.append(page)
            else:
                raise ValueError(
                    "Pages section is malformed, refer to sample config.toml"
                )

        # Validate pages
        # Page space may be none, in that case default space must be specified
        # Page name may be none, but the pages list may contain only one. In that case,
        # page title is expected in the main script
        for page in self.__pages:
            if page.page_space is None:
                if default_space is not None:
                    page.page_space = default_space
                else:
                    raise ValueError(
                        f"Page '{page.page_title}' does not have page_space specified,"
                        f" neither is default space"
                    )
            if page.page_title is None and len(self.__pages) > 1:
                raise ValueError(
                    "There are more than 1 page, and one of the names is not specified"
                )

            # Check that there are no pages with same space and name - they will overwrite each other
            page_title_func = attrgetter("page_title")
            page_space_func = attrgetter("page_space")
            groups = groupby(sorted(self.__pages, key=page_title_func), page_title_func)
            for page_title, g in groups:
                _ = list(g)
                groups_space = groupby(sorted(_, key=page_space_func), page_space_func)
                for space_name, group in groups_space:
                    _ = list(group)
                    if len(_) > 1:
                        raise ValueError(
                            f"There are more than 1 page called '{page_title}' in space {space_name}"
                        )

    @property
    def auth(self):
        return self.__auth

    @auth.setter
    def auth(self, auth: dict):
        for mandatory_config in ["confluence_url", "username", "is_cloud"]:
            if mandatory_config not in auth:
                raise KeyError(f"{mandatory_config} not in auth section")
        self.__auth = Auth(
            auth["confluence_url"],
            auth["username"],
            auth.get("password", None),
            auth["is_cloud"],
        )

    @property
    def author(self):
        return self.__author

    @author.setter
    def author(self, author: Union[str, None]):
        if author is None:
            author = self.auth.username
        if not isinstance(author, str):
            raise ValueError("Author is not a string")
        elif len(author) == 0:
            raise ValueError(
                "Author's name is specified in the config but is empty. "
                "Specify it or remove the setting to use the authentication username for checks."
            )

        self.__author = author



import typer
import sys
from click import Choice
from typing import Optional, List, Tuple
from pathlib import Path
from logging import basicConfig, DEBUG
from atlassian import Confluence
from atlassian.errors import ApiError
from dataclasses import dataclass, field, astuple
from requests.exceptions import ConnectionError

from confluence_poster.poster_config import Page, AllowedFileFormat
from confluence_poster.config_loader import load_config
from confluence_poster.config_wizard import DialogParameter, generate_page_dialog_params
from confluence_poster.main_helpers import (
    check_last_updated_by,
    PostedPage,
    StateConfig,
    get_page_url,
)
from confluence_poster.convert_utils import (
    guess_file_format,
    get_representation_for_format,
    post_to_convert_api,
    convert_using_markdown_lib,
)
from confluence_poster.page_creation_helpers import create_page
from confluence_poster.file_upload_helpers import attach_files_to_page

__version__ = "1.4.0"
default_config_name = "config.toml"


def version_callback(value: bool):
    if value:
        typer.echo(f"Confluence poster version: {__version__}")
        raise typer.Exit()


@dataclass
class Report:
    created_pages: List[Page] = field(default_factory=list)
    updated_pages: List[Page] = field(default_factory=list)
    unprocessed_pages: List[Tuple[Page, str]] = field(default_factory=list)
    confluence_instance: Confluence = None

    def __str__(self) -> str:
        output = ""
        for header, page_list in [
            ("Created pages:", self.created_pages),
            ("Updated pages:", self.updated_pages),
        ]:
            output += header + "\n"
            if page_list:
                for page in page_list:
                    title = page.page_title
                    space = page.page_space
                    output += f"{space}::{title} {get_page_url(title, space, self.confluence_instance)}\n"
            else:
                output += "None\n"
        if self.unprocessed_pages:
            output += "Unprocessed pages:"
            for page, reason in self.unprocessed_pages:
                output += f"{page.page_space}::{page.page_title} Reason: {reason}"

        return output


app = typer.Typer()
state = StateConfig()


@app.command()
def convert_markdown(
    use_confluence_converter: Optional[bool] = typer.Option(
        False,
        "--use-confluence-converter",
        show_default=False,
        help="Use built-in Confluence converter. Note: uses Confluence private API.",
    )
):
    """Converts single page text to html. Prints the converted text.
    Implies running the utility with --quiet. Logs runtime info only to stderr.

    When the page is posted, "editor" representation should be used.

    By default uses python's markdown library with fenced code and tables extensions to render markdown to html.

    If --use-confluence-converter flag is used - uses Confluence built-in converter."""
    confluence = state.confluence_instance
    always_echo = state.always_print_function
    echo_err = state.print_stderr
    text = state.config.pages[0].page_text

    if len(state.config.pages) > 1:
        echo_err("This command supports converting only one page at a time.")
        raise typer.Exit(1)

    if use_confluence_converter:
        echo_err(
            "Using the converter built into Confluence which is labeled as private API. "
            "The results may be less than satisfactory."
        )
        always_echo(post_to_convert_api(confluence, text))
    else:
        always_echo(convert_using_markdown_lib(text))

    echo_err(
        "Submit the converted text using `confluence_poster post-page --file-format html`.",
    )


@app.command()
def post_page(
    upload_files: Optional[bool] = typer.Option(
        False, "--upload-files", show_default=False, help="Upload list of files."
    ),
    version_comment: Optional[str] = typer.Option(
        None,
        "--version-comment",
        show_default=False,
        help="Provider version comment.",
    ),
    create_in_space_root: Optional[bool] = typer.Option(
        False,
        "--create-in-space-root",
        show_default=False,
        help="Create the page in space root.",
    ),
    file_format: Optional[AllowedFileFormat] = typer.Option(
        AllowedFileFormat.none,
        "--file-format",
        show_default=False,
        help="File format of the file with the page content. "
        "If provided at runtime - can only be applied to a single page. "
        "If set to 'None'(default) - script will try to guess it during the run.",
    ),
    files: Optional[List[Path]] = typer.Argument(None, help="List of files to upload"),
):
    """Posts the content of the pages."""
    echo = state.print_function
    always_echo = state.always_print_function
    echo_err = state.print_stderr
    confirm = state.confirm_function
    prompt = state.prompt_function

    report = Report(confluence_instance=state.confluence_instance)
    confluence = state.confluence_instance
    posted_pages = [PostedPage(*astuple(_)) for _ in state.config.pages]
    target_page = posted_pages[0]

    if len(posted_pages) > 1 and version_comment is not None:
        apply_version_comment_to = prompt(
            text=f"Multiple pages specified. Do you want to apply the comment to [A]ll pages, "
            "[F]irst one or [N]ot apply it?",
            type=Choice(choices=["A", "F", "N"], case_sensitive=False),
            default="A",
        ).lower()
        if apply_version_comment_to == "a":
            for page in posted_pages:
                page.version_comment = version_comment
        elif apply_version_comment_to == "f":
            posted_pages[0].version_comment = version_comment
        else:
            for page in posted_pages:
                page.version_comment = None
    else:
        # set the comment for the single page
        posted_pages[0].version_comment = version_comment

    # file format check
    if len(posted_pages) > 1 and file_format is not AllowedFileFormat.none:
        echo_err(
            "File format cannot be used for cases when there are more than 1 pages in the config. "
            "Consider adding it to the config file."
        )
        raise typer.Exit(1)

    posted_pages[0].page_file_format = file_format

    if upload_files:
        if len(posted_pages) > 1:
            always_echo(
                "Upload files are specified, but there are more than 1 pages in the config."
            )
            if confirm(
                f"Continue by attaching all files to the first page, '{target_page.page_title}'? ('N') to abort",
                default=False,
            ):
                pass
            else:
                echo_err("Aborting.")
                raise typer.Exit(3)

    for page in posted_pages:
        if page.page_file_format is AllowedFileFormat.none:
            echo(
                f"File format for page {page.page_title} not specified. Trying to determine it..."
            )
            try:
                guessed_format = guess_file_format(page.page_file)
            except ValueError as e:
                echo_err(
                    "Could not guess the file format. Consider specifying it manually. "
                    "See --help for information.",
                )
                raise e
            echo(f"Guessed file format as {guessed_format.value}")
            page.page_file_format = guessed_format

        if page.page_file_format is AllowedFileFormat.markdown:
            page.page_text = convert_using_markdown_lib(page.page_text)

        echo(f"Looking for page '{page.page_title}'")
        if page_id := confluence.get_page_id(
            space=page.page_space, title=page.page_title
        ):
            # Page exists
            echo(f"Found page id #{page_id}")
            page.page_id = page_id

            # If --force is supplied - we do not really care about who edited the page last
            if not (state.force or page.force_overwrite):
                updated_by_author, page_last_updated_by = check_last_updated_by(
                    page_id=page_id,
                    username_to_check=state.config.author,
                    confluence_instance=confluence,
                )
                if not updated_by_author:
                    echo(
                        f"Flag 'force' is not set and last author of page '{page.page_title}'"
                        f" is {page_last_updated_by}, not {state.config.author}. Skipping page"
                    )
                    continue
            else:
                if state.force:
                    echo("Flag 'force' set globally.")
                elif page.force_overwrite:
                    echo("Flag 'force overwrite' set on the page.")
                echo("Author name check skipped.")

            echo(f"Updating page #{page_id}")
            confluence.update_existing_page(
                page_id=page_id,
                title=page.page_title,
                body=page.page_text,
                representation=get_representation_for_format(
                    page.page_file_format
                ).value,
                minor_edit=state.minor_edit,
                version_comment=page.version_comment,
            )
            report.updated_pages += [page]
        else:
            echo(
                f"Could not find page '{page.page_title}' in space '{page.page_space}'"
            )
            if page_created := create_page(
                page=page, state=state, create_in_root=create_in_space_root
            ):
                report.created_pages += [page]
                if version_comment:
                    echo(
                        "Page was created, but Confluence API does not support setting the version comment for"
                        " page creation. The comment was not saved in the page history."
                    )
                page.page_id = page_created.page_id
            else:
                always_echo(f"Not creating page '{page.page_title}'")
                report.unprocessed_pages += [(page, page_created.comment)]

        if upload_files and target_page.page_id is not None:
            attach_files_to_page(page=target_page, files=files, state=state)

    always_echo("Finished processing pages")

    if state.print_report:
        always_echo(report)


@app.command()
def validate(
    online: Optional[bool] = typer.Option(
        False,
        "--online",
        show_default=False,
        help="Test the provided authentication settings on the actual"
        " instance of Confluence.",
    )
):
    """Validates the provided settings. If 'online' flag is passed - tries to fetch the space from the config using the
    supplied credentials."""
    echo = state.print_function
    echo_err = state.print_stderr

    if online:
        echo("Validating settings against the Confluence instance from config")
        try:
            space_key = state.config.pages[0].page_space
            state.print_function(f"Trying to get {space_key}...")
            space_id = state.confluence_instance.get_space(space_key).get("id")
        except ConnectionError:
            echo_err(
                f"Could not connect to {state.config.auth.url}. Make sure it is correct",
            )
            raise typer.Abort(1)
        except ApiError as e:
            echo_err(f"Got an API error, details: {e.reason}")
            raise typer.Abort(1)
        else:
            echo(f"Got space id #{space_id}.")
    echo("Validation successful")


@app.command()
def create_config(
    local_only: Optional[bool] = typer.Option(
        False,
        "--local-only",
        show_default=False,
        help="Create config only in the local folder.",
    ),
    home_only: Optional[bool] = typer.Option(
        False,
        "--home-only",
        show_default=False,
        help="Create config only in the $XDG_CONFIG_HOME.",
    ),
):
    """Runs configuration wizard. The wizard guides through setting up values for configuration file."""
    import xdg
    from confluence_poster.config_wizard import (
        config_dialog,
        get_filled_attributes_from_file,
        print_config_with_hidden_attrs,
        page_add_dialog,
    )
    from functools import partial

    echo = state.print_function
    confirm = state.confirm_function
    prompt = state.prompt_function

    home_config_location = xdg.xdg_config_home() / "confluence_poster/config.toml"

    all_params = (
        DialogParameter(
            "author",
            comment="If the page was not updated by the username specified here, throw an error."
            "\nIf this setting is omitted - username from auth section "
            "is used for checks",
            required=False,
        ),
        # auth:
        DialogParameter("auth.confluence_url", comment="URL of confluence instance"),
        DialogParameter(
            "auth.username", comment="Username for authentication in Confluence"
        ),
        DialogParameter(
            "auth.password",
            comment="Password for authentication. May be supplied through runtime option or "
            "environment",
            required=False,
            hide_input=True,
        ),
        DialogParameter(
            "auth.is_cloud",
            comment="Whether the confluence instance is a cloud one",
            type=bool,
        ),
        # pages:
        DialogParameter(
            "pages.default.page_space",
            comment="Space key (e.g. LOC for 'local-dev' space). If defined here - will be used "
            "if a page does not redefine it",
            required=False,
        ),
    ) + generate_page_dialog_params(1)
    home_only_params = (
        "author",
        "auth.confluence_url",
        "auth.username",
        "auth.password",
        "auth.is_cloud",
    )
    # To hide password in prompts
    _print_config_file = partial(
        print_config_with_hidden_attrs, hidden_attributes=["auth.password"]
    )
    config_dialog = partial(config_dialog, config_print_function=_print_config_file)
    page_add_dialog = partial(page_add_dialog, config_print_function=_print_config_file)

    # Initial prompt
    echo("Starting config wizard.")
    echo("This wizard will guide you through creating the configuration files.")
    answer = ""
    if not any([local_only, home_only]):
        echo(
            "Since neither '--local-only' nor '--home-only' were specified, wizard will guide you through creating "
            f"config files in {home_config_location.parent}(XDG_CONFIG_HOME) and {Path.cwd()}(local directory)"
        )

        answer = prompt(
            f"Create config in {home_config_location.parent}? [Y/n/q]"
            "\n* 'n' skips to config in the local directory"
            "\n* 'q' will exit the wizard\n",
            type=str,
            default="y",
        ).lower()

        if answer == "q":
            raise typer.Exit()

    if (answer == "y" and not local_only) or home_only:
        # Create config in home
        while True:
            dialog_result = config_dialog(
                filename=home_config_location,
                attributes=[_ for _ in all_params if _ in home_only_params],
            )
            if dialog_result is None or dialog_result:
                # None means the user does not want to overwrite the file
                break

    if home_only:
        # If --home-only is specified - no need to create another one in local folder
        echo("--home-only specified, not attempting to create any more configs.")
        raise typer.Exit()

    if not local_only:
        # If local-only is passed - no need to ask for confirmation of creating a local only config
        local_answer = confirm(
            f"Proceed to create config in {Path.cwd()}?", default=True
        )
        if not local_answer:
            echo("Exiting.")
            raise typer.Exit()

    # Create config in current working directory
    echo("Creating config in local directory.")
    local_config_name = prompt(
        "Please provide the name of local config", type=str, default=default_config_name
    )

    home_parameters = get_filled_attributes_from_file(home_config_location)
    local_config_parameters = [_ for _ in all_params if _ not in home_parameters]

    while True:
        dialog_result = config_dialog(
            filename=Path.cwd() / local_config_name, attributes=local_config_parameters
        )
        if dialog_result is None or dialog_result:
            # None means the user does not want to overwrite the file
            break

    while confirm("Add more pages?", default=False):
        page_add_dialog(Path.cwd() / local_config_name)

    echo(
        "Configuration wizard finished. Consider running the `validate` command to check the generated config"
    )


@app.callback()
def main(
    ctx: typer.Context,
    version: Optional[bool] = typer.Option(
        None, "--version", help="Show version and exit.", callback=version_callback
    ),
    config: Path = typer.Option(
        default=default_config_name,
        help="The file containing configuration. "
        f"If not specified - {default_config_name} from the same directory is used",
    ),
    page_title: Optional[str] = typer.Option(
        None,
        help="Override page title from config."
        " Applicable if there is only one page.",
    ),
    parent_page_title: Optional[str] = typer.Option(
        None,
        help="Provide a parent title to search for."
        " Applicable if there is only one page.",
    ),
    page_file: Optional[Path] = typer.Option(
        None,
        help="Provide the path to the file containing page text. Allows passing '-' to read from stdin.",
    ),
    password: Optional[str] = typer.Option(
        None, help="Supply the password in command line.", envvar="CONFLUENCE_PASSWORD"
    ),
    force: Optional[bool] = typer.Option(
        False,
        "--force",
        show_default=False,
        help="Force overwrite the pages. Skips all checks for different author of the updated page. "
        "To set for individual pages you can specify field 'force_overwrite' in config.",
    ),
    force_create: Optional[bool] = typer.Option(
        False,
        "--force-create",
        show_default=False,
        help="Disable prompts to create pages. "
        "Script could still prompt for a parent page.",
    ),
    minor_edit: Optional[bool] = typer.Option(
        False,
        "--minor-edit",
        show_default=False,
        help="Do not notify watchers of pages updates. " "Not enabled by default.",
    ),
    report: Optional[bool] = typer.Option(
        False,
        "--report",
        show_default=False,
        help="Print report at the end of the run. " "Not enabled by default.",
    ),
    debug: Optional[bool] = typer.Option(
        False,
        "--debug",
        show_default=False,
        help="Enable debug logging. " "Not enabled by default.",
    ),
    quiet: Optional[bool] = typer.Option(
        False,
        "--quiet",
        show_default=False,
        help="Suppresses certain output.",
    ),
):
    """Supplementary script for writing Confluence articles in
    local editor. Uses information from the config to post the article content to confluence.

    """

    if ctx.invoked_subcommand == "convert-markdown":
        quiet = True

    if str(page_file) == "-":
        state.filter_mode = True
        if ctx.invoked_subcommand not in {"convert-markdown", "post-page"}:
            typer.echo(
                f"Invoked command, {ctx.invoked_subcommand} is not compatible with reading page text from stdin",
                err=True,
            )
            raise typer.Exit(3)
    else:
        state.filter_mode = False

    state.quiet = quiet

    echo = state.print_function
    always_echo = state.always_print_function
    echo_err = state.print_stderr

    echo("Starting confluence_poster")
    if debug:
        from pprint import pprint

        always_echo("Running in debug mode.")

        echo("Set options:")
        state.debug = True
        echo(pprint(locals()))
        # Set global debug to true, for other modules
        basicConfig(level=DEBUG, format="%(asctime)s %(levelname)s %(message)s")
    else:
        state.debug = False

    if (
        ctx.invoked_subcommand != "create-config"
    ):  # no need to validate or load the config if we're creating it
        state.force = force
        state.force_create = force_create
        state.print_report = report
        state.minor_edit = minor_edit

        echo("Reading config")
        try:
            confluence_config = load_config(config)
        except FileNotFoundError as e:
            echo_err("Config file not found. Consider running `create-config`")
            raise e
        state.config = confluence_config

        # Check that the parameters are not used with more than 1 page in the config
        if page_title or parent_page_title or page_file:
            if len(confluence_config.pages) > 1:
                echo_err(
                    "Page title, parent page title or page file specified as a parameter "
                    "but there are more than 1 page in the config.\n"
                    "These parameters are intended to be used with only one page.\n"
                    "Please specify them in the config.\n"
                    "Aborting.",
                )
                raise typer.Exit(1)
            if page_title:
                state.config.pages[0].page_title = page_title
            if parent_page_title:
                state.config.pages[0].parent_page_title = parent_page_title
            if page_file:
                if state.filter_mode:
                    state.config.pages[0].page_text = sys.stdin.read()
                else:
                    state.config.pages[0].page_file = page_file

        # Validate password
        try:
            _password = next(
                _ for _ in [password, confluence_config.auth.password] if _ is not None
            )
        except StopIteration:
            echo_err(
                "Password is not specified in environment, parameter or the config. Aborting",
            )
            raise typer.Exit(1)

        # set API version
        if confluence_config.auth.is_cloud:
            api_version = "cloud"
        else:
            api_version = "latest"

        state.confluence_instance = Confluence(
            url=confluence_config.auth.url,
            username=confluence_config.auth.username,
            password=_password,
            api_version=api_version,
        )



from atlassian import Confluence
from dataclasses import dataclass, field
from typing import Union, Callable, List
from typer import echo, prompt, confirm
from functools import partial

from confluence_poster.poster_config import Page, Config

"""File that contains procedures used inside main.py's functions"""


def get_page_url(
    page_title: str, space: str, confluence: Confluence
) -> Union[str, None]:
    """Retrieves page URL"""
    if page := confluence.get_page_by_title(space=space, title=page_title, expand=""):
        # according to Atlassian REST API reference, '_links' is a legitimate way to access links
        page_link = confluence.url + page["_links"]["webui"]
        return page_link
    else:
        return None


def check_last_updated_by(
    page_id: int, username_to_check: str, confluence_instance: Confluence
) -> (bool, str):
    """Checks which user last updated `page_id`. If it's not `username_to_check` — return False
    :param page_id: ID of the page to check
    :param username_to_check: compare this username against the one that last updated the page
    :param confluence_instance: instance of Confluence to run the check
    """
    page_last_updated_by = confluence_instance.get_page_by_id(
        page_id, expand="version"
    )["version"]["by"]
    if confluence_instance.api_version == "cloud":
        page_last_updated_by = page_last_updated_by["email"]  # pragma: no cover
    else:
        page_last_updated_by = page_last_updated_by["username"]

    return page_last_updated_by == username_to_check, page_last_updated_by


@dataclass
class PostedPage(Page):
    """Merges independently set fields with the runtime-set fields"""

    version_comment: Union[str, None] = None
    page_id: Union[int, None] = None


@dataclass
class StateConfig:
    """Holds the shared state between typer commands"""

    force: bool = False
    debug: bool = False
    confluence_instance: Union[None, Confluence] = None
    config: Union[None, Config] = None
    minor_edit: bool = False
    print_report: bool = False
    force_create: bool = False
    created_pages: List[int] = field(default_factory=list)
    _filter_mode: bool = False
    quiet: bool = False

    @property
    def print_function(self) -> Callable:
        if self.quiet:
            # noinspection PyUnusedLocal
            def _quiet_func(*args, **kwargs):
                pass

            return _quiet_func
        else:
            return echo

    @property
    def always_print_function(self) -> Callable:
        return echo

    @property
    def prompt_function(self) -> Callable:
        if self.filter_mode:
            # noinspection PyUnusedLocal
            def _raise_exception(*args, **kwargs):
                raise Exception("Prompt invoked in filter mode!")

            return _raise_exception
        else:
            return prompt

    @property
    def confirm_function(self) -> Callable:
        if self.filter_mode:
            # noinspection PyUnusedLocal
            def _raise_exception(*args, **kwargs):
                # TODO: show prompt?
                raise Exception("Confirmation prompt invoked in filter mode!")

            return _raise_exception
        else:
            return confirm

    @property
    def print_stderr(self) -> Callable:
        return partial(echo, err=True)

    @property
    def filter_mode(self):
        return self._filter_mode

    @filter_mode.setter
    def filter_mode(self, value: bool):
        self._filter_mode = value
        if value:
            self.quiet = True



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

setup(
    name="confluence_poster",
    version="1.4.0",
    description="Script that updates Confluence articles from files written locally",
    long_description=(here / "README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    url="https://github.com/VTimofeenko/confluence_poster",
    license="MIT",
    author="Vladimir Timofeenko",
    author_email="confluence.poster.maintain@vtimofeenko.com",
    include_package_data=True,
    package_data={"confluence_poster": ["config.toml"]},
    packages=find_packages(exclude=("tests", "docs")),
    package_dir={"confluence_poster": "confluence_poster"},
    entry_points={
        "console_scripts": ["confluence_poster = confluence_poster.main:app"]
    },
    install_requires=[
        "atlassian-python-api==2.3.0",
        "typer>=0.3.2",
        "toml",
        "requests",
        "xdg>=5.0.1",
        "tomlkit==0.7.0",
        "click>=7.1.1, <7.2.0",
        "markdown==3.3.3",
        "marshmallow==3.10.0",
    ],
    python_requires=">=3.8, <4",
    tests_require=[
        "pytest",
        "pytest-cov",
        "faker",
        "beautifulsoup4==4.9.3",
        "bs4",
        "lxml",
    ],
    extras_require={"docs": ["jinja2", "typer-cli"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Topic :: Utilities",
        "Topic :: Text Processing",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
    ],
    keywords="confluence api vim",
    project_urls={
        "Bug Reports": "https://github.com/VTimofeenko/confluence_poster/issues",
        "Source": "https://github.com/VTimofeenko/confluence_poster/",
    },
)
